% Main program
% Run Initialization file first

%%%%%%%%%%%%%%%%
% Get Variables
%%%%%%%%%%%%%%%%
% disturbances 
%valve opening [%] 
cv101 = P_vector(1);
cv102 = P_vector(2);
cv103 = P_vector(3);
% if value is [A] from 0.004 to 0.020
% if you want to convert to 0 (fully closed) to 1 (fully open)
% vo_n = (vo - 0.004)./(0.02 - 0.004);

%pump rotation [%]
pRate = P_vector(4);
% if value is [A] from 0.004 to 0.020
% if you want to convert to (min speed - max speed)
% goes from 12% of the max speed to 92% of the max speed
% pRate = 12 + (92 - 12)*(P_vector(4) - 0.004)./(0.02 - 0.004);

% always maintain the inputs greater than 0.5
% inputs computed in the previous MPC iteration
% Note that the inputs are the setpoints to the gas flowrate PID's
fic104sp = P_vector(5);
fic105sp = P_vector(6); 
fic106sp = P_vector(7);
%current inputs of the plant
u0old=[P_vector(5);P_vector(6);P_vector(7)];


% cropping the data vector
nd = size(I_vector,2);
dataCrop = (nd - BufferLength + 1):nd;

% liquid flowrates [L/min] %INCLUDE NOISE???
fi101 = I_vector(1,dataCrop);
fi102 = I_vector(2,dataCrop);
fi103 = I_vector(3,dataCrop);

% actual gas flowrates [sL/min] %INCLUDE NOISE???
fic104 = I_vector(4,dataCrop);
fic105 = I_vector(5,dataCrop);
fic106 = I_vector(6,dataCrop);

% pressure @ injection point [mbar g]
pi105 = I_vector(7,dataCrop);
pi106 = I_vector(8,dataCrop);
pi107 = I_vector(9,dataCrop);

% reservoir outlet temperature [oC]
ti101 = I_vector(10,dataCrop);
ti102 = I_vector(11,dataCrop);
ti103 = I_vector(12,dataCrop);

% DP @ erosion boxes [mbar]
dp101 = I_vector(13,dataCrop);
dp102 = I_vector(14,dataCrop);
dp103 = I_vector(15,dataCrop);

% top pressure [mbar g]
% for conversion [bar a]-->[mbar g]
% ptop_n = ptop*10^-3 + 1.01325;
pi101 = I_vector(16,dataCrop);
pi102 = I_vector(17,dataCrop);
pi103 = I_vector(18,dataCrop);

% reservoir pressure [bar g]
% for conversion [bar g]-->[bar a]
% ptop_n = ptop + 1.01325;
pi104 = I_vector(19,dataCrop);

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CODE GOES HERE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% number of measurements in the data window
dss = size(pi104,2);

% we do no consider SS detection here
flagSS = 1;

% check for first iteration
if ~exist('dxHatkk','var') 
    %initial condition
    [dx0,z0,u0,theta0] = InitialConditionGasLift(par);

    dxHatkk = dx0;
    zHatkk = z0;
    thetakk = theta0;

    load('EKFconf');   
    qThres = ekf.qThresk;
    Pkk = ekf.Pkk;
    %ekf.R = noise.output; %added (different from Rig Implementation)
    
    lambda = 9.5;
    err0 = zeros(3,1);
end

if flagSS == 1
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Estimating SS model parameters (dynamic) %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    yPlant = [fi101; 
              fi102;
              fi103;
              1.01325 + 10^-3*pi101; %[mbarg]-->[bar a];
              1.01325 + 10^-3*pi102;
              1.01325 + 10^-3*pi103];
          
    uPlant = [fic104; %conversion [L/min] --> [kg/s]
              fic105;
              fic106;
              cv101*ones(1,dss); %workaround - i just have the last measurement here. Since it is the disturbance, it doesn't really matter;
              cv102*ones(1,dss);
              cv103*ones(1,dss);
              pi104 + 1.01325];            
    
          try
              % running casadi - getting the last measurement and last
              % input that generated that measuremnt
              [dxHatkk,zHatkk,thetakk,Pkk,qThres] = RExtendedKalmanFilter(yPlant(:,end),uPlant(:,end - 10),dxHatkk,zHatkk,thetakk,Pkk,F_model,S_xx,S_zz,S_xz,S_xp,S_zp,ekf,par);
              
              % everything normal
              flagEst = 1;
          catch
              warning('Dynamic Estimation Problem!');
              beep
              % estimation problem
              flagEst = 0;
              
              % Note that in this case, we dont update
              % dxHatkk,zHatkk,thetakk,Pkk,qThres
          end
    
    
    if flagEst == 1
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Optimizing Systems %
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % Update Lambda
        Total_Qgl = fic104(end) + fic105(end) + fic106(end);
      
%         K_ccc = 1.2060;
        K_ccc = 0.907;
%         tau_ccc = 15; %284/4;
        tau_ccc = 125;
        theta_ccc = 0;
        KP_ccc = 1/(K_ccc*(tau_ccc+theta_ccc));
        KI_ccc = 1/(K_ccc*(tau_ccc+theta_ccc));
        % lambda_next =  max(0,lambda + 0.75 * (Total_Qgl   - 7.5));
        
        tStart = cputime;
        
        lambda_next =  max(0,lambda + KI_ccc*(Total_Qgl   - 7.5));% + KP_ccc*(Total_Qgl   - 7.5)); 
%         lambda_next = lambda;
        
        tEnd_lambda = cputime - tStart;
        
        lambda = lambda_next;

%          if kk>200
%             lambda = 10.0023;
%          end
        
        
        % Gradient Estimation
        %conversion
        CR = 60*10^3; % [L/min] -> [m3/s] 
        
        %%% NEED TO CHECK
        F_zp = full(S_zp(dxHatkk,zHatkk,[uPlant(:,end);thetakk;1e4]));
        
        tStart = cputime;
        %F_zp = full(S_zp(dxHatkk,zHatkk,[uSPPlantArray(1:3,end);uPlant(4:end,end);thetakk;1e4]));
        
        %F_zp = full(S_zp(dxHatkk,zHatkk,[uPlant(1:end,end);thetakk;1e4]));
        tEnd_lambda = cputime - tStart + tEnd_lambda;
        %%% NEED TO CHECK
        
%         %states
%         F_xx = full(S_xx(dxHatkk,zHatkk,[uPlant(:,end - 10);thetakk;par.T]));
%         F_zz = full(S_zz(dxHatkk,zHatkk,[uPlant(:,end - 10);thetakk;par.T]));
%         F_xz = full(S_xz(dxHatkk,zHatkk,[uPlant(:,end - 10);thetakk;par.T]));
%         F_zx = F_xz';    
% 
%         A = [F_xx, F_xz; F_zx, F_zz];
% 
%         %inputs
%         F_xp = full(S_xp(dxHatkk,zHatkk,[uPlant(:,end - 10);thetakk;par.T]));
%         F_zp = full(S_zp(dxHatkk,zHatkk,[uPlant(:,end - 10);thetakk;par.T]));
%         
%         B = [F_xp(:,1:3); F_zp(:,1:3)];
%         
%         C = zeros(1,30);
%         C(1,28)=- 20*((1*1e-2)*CR/par.rho_o(1));
%         C(1,29)=- 25*((1*1e-2)*CR/par.rho_o(2));
%         C(1,30)=- 30*((1*1e-2)*CR/par.rho_o(3));
%         
%         Ju = (-C*(A\B))';

        J_qgl1 = - 20*((1*1e-2)*CR/par.rho_o(1))*F_zp(22,1);
        J_qgl2 = - 25*((1*1e-2)*CR/par.rho_o(2))*F_zp(23,2);
        J_qgl3 = - 30*((1*1e-2)*CR/par.rho_o(3))*F_zp(24,3);

%         J_qgl1 = Ju(1);
%         J_qgl2 = Ju(2);
%         J_qgl3 = Ju(3);        

        g_qgl1 = 1;
        g_qgl2 = 1;
        g_qgl3 = 1;
        
        L_qgl1 = J_qgl1 + lambda.*g_qgl1;
        L_qgl2 = J_qgl2 + lambda.*g_qgl2;
        L_qgl3 = J_qgl3 + lambda.*g_qgl3;
        
        Lu(:,i) = [L_qgl1;L_qgl2;L_qgl3];
        
        %tauC = 3000;
        %tauC = 3000;
        tauC = 10;
        %tauI = min(4*tauC,800);
        %tauI = min(4*tauC,1);
        tauI = min(4*tauC,800);

        % --------- Well 1 ---------
        %Kp1 = 1*200/(0.439*tauC);
        %Kp1 = 0.5*200/(0.439*tauC);
        K_1 = 2.53;
%         tauC_1 = 2;%26/2;
        tauC_1 = 25;
        theta_1 = 0;
        Kp1 = 1/K_1;
        Ki1 = 1/(K_1*tauC_1 + theta_1);
%         Ki1 = Kp1/tauI;
        err_1= -L_qgl1;
            %uOpt(1,1) = max(1,min(5.5, u0old(1) + (Kp1*err_1 + Ki1*err_1 - 0*Kp1*err0(1))));
            % uOpt(1,1) = max(1,min(5.5, u0old(1) + (Kp1*err_1 + Ki1*err_1 - 1*Kp1*err0(1))));
            % uOpt(1,1) = max(1,min(5.5, u0old(1) + (Kp1*err_1 + 0*Ki1*err_1 - 0*Kp1*err0(1))));
        
        tStart = cputime;
                uOpt(1,1) = max(1,min(5.5, u0old(1) + (Ki1*err_1))); %+ (Kp1*err_1)
        tEnd_gr1 = cputime - tStart;        
        
%                 if kk>75
%                    uOpt(1,1) = 3;
%                 end
        err0(1) = err_1;
        % --------- Well 2 ---------
        %Kp1 = 1*150/(0.29*tauC);
        %Kp1 = 0.5*100/(0.29*tauC);
        K_2 = 2.93;
%         tauC_2 = 2;%16/2;
        tauC_2 = 25;
        theta_2 = 0;
        Kp2 = 1/K_2;
        Ki2 = 1/(K_2*tauC_2 + theta_2);
%        Ki1 = Kp1/tauI;
        err_2 = -L_qgl2;
            %uOpt(2,1) = max(1,min(5.5, u0old(2) + (Kp1*err_2 + Ki1*err_2 - 0*Kp1*err0(2)))); 
            %uOpt(2,1) = max(1,min(5.5, u0old(2) + (Kp1*err_2 + Ki1*err_2 - 1*Kp1*err0(2))));  
            % uOpt(2,1) = max(1,min(5.5, u0old(2) + (Kp1*err_2 + 0*Ki1*err_2 - 0*Kp1*err0(2))));  
            tStart = cputime;
                uOpt(2,1) = max(1,min(5.5, u0old(2) + (Ki2*err_2))); %+ (Kp2*err_2)
            tEnd_gr2 = cputime - tStart; 
%                 if kk>75
%                    uOpt(2,1) = 3;
%                 end
        err0(2) = err_2;
         % --------- Well 3 ---------
        %Kp1 = 1*250/(0.439*tauC);
        %Kp1 = 0.5*250/(0.439*tauC);
        K_3 = 3.70;
%         tauC_3 = 2;%14/2;
        tauC_3 = 25;%14/2;
        theta_3 = 0;
        Kp3 = 1/K_3;
        Ki3 = 1/(K_3*tauC_3 + theta_3);
        %Ki1 = Kp1/tauI;
        err_3= -L_qgl3;
            %uOpt(3,1) = max(1,min(5.5, u0old(3) + (Kp1*err_3 + Ki1*err_3 - 0*Kp1*err0(3))));
            %uOpt(3,1) = max(1,min(5.5, u0old(3) + (Kp1*err_3 + Ki1*err_3 - 1*Kp1*err0(3))));
            % uOpt(3,1) = max(1,min(5.5, u0old(3) + (Kp1*err_3 + 0*Ki1*err_3 - 0*Kp1*err0(3))));
            tStart = cputime;
                 uOpt(3,1) = max(1,min(5.5, u0old(3) + (Ki3*err_3)));% + (Kp3*err_3)
            tEnd_gr3 = cputime - tStart;
%                 if kk>75
%                    uOpt(3,1) = 3;
%                 end
        err0(3) = err_3;        
        
%         uOpt=[u0old(1) u0old(2) 2.4003]';
        flagOpt = 1;
        
        tEnd = max(max(tEnd_gr1,tEnd_gr2),tEnd_gr3)+ tEnd_lambda;
        % getting mean value of the inputs as initial guess for optimization
        %uGuess = mean(uPlant,2);
        %uGuess(1:3) = [2.5; 2.5; 2.5];
        
        %[uOpt,xOptHat,zOptHat,phi,flagOpt] = ErosionRigSSOptimization(dxHatkk,zHatkk,thetakk,uGuess,par);

        if flagOpt == 1
            %%%%%%%%%%%%%%%%%%%%%%%%%%%
            %Send Variable
            %%%%%%%%%%%%%%%%%%%%%%%%%%%
            % compute new values for the gas flow rate setpoints
            %Filter for optimum
            fic104sp = uOpt(1);
            fic105sp = uOpt(2);
            fic106sp = uOpt(3);

            O_vector = vertcat(fic104sp,fic105sp,fic106sp)';
            SS = 1;
            Estimation = 1;
            Optimization = 1;
            Result = lambda; 
            Parameter_Estimation = thetakk';
            State_Variables_Estimation = (par.H*zHatkk)';
            State_Variables_Optimization = ([J_qgl1;J_qgl2;J_qgl3;L_qgl1;L_qgl2;L_qgl3;])';
            Optimized_Air_Injection = uOpt';
        else
            %%%%%%%%%
            %(dummy)%
            %%%%%%%%%
            % compute new values for the gas flow rate setpoints
            O_vector = vertcat(fic104sp,fic105sp,fic106sp)';

            SS = 1;
            Estimation = 1;
            Optimization = 0;
            Result = 0;
            Parameter_Estimation = [0,0,0,0,0,0];
            State_Variables_Estimation = [0,0,0,0,0,0];
            State_Variables_Optimization = [0,0,0,0,0,0];
            Optimized_Air_Injection = [0,0,0];
        end
        
    else  
        %%%%%%%%%
        %(dummy)%
        %%%%%%%%%
        % compute new values for the gas flow rate setpoints
        O_vector = vertcat(fic104sp,fic105sp,fic106sp)';

        SS = 1;
        Estimation = 0;
        Optimization = 0;
        Result = 0;
        Parameter_Estimation = [0,0,0,0,0,0];
        State_Variables_Estimation = [0,0,0,0,0,0];
        State_Variables_Optimization = [0,0,0,0,0,0];
        Optimized_Air_Injection = [0,0,0];
    end
    
    
else
    %%%%%%%%%
    %(dummy)%
    %%%%%%%%%
    % compute new values for the gas flow rate setpoints
    O_vector = vertcat(fic104sp,fic105sp,fic106sp)';
    
    SS = 0;
    Estimation = 0;
    Optimization = 0;
    Result = 0;
    Parameter_Estimation = [0,0,0,0,0,0];
    State_Variables_Estimation = [0,0,0,0,0,0];
    State_Variables_Optimization = [0,0,0,0,0,0];
    Optimized_Air_Injection = [0,0,0];
    
end
