% Main program
% Run Initialization file first

%%%%%%%%%%%%%%%%
% Get Variables
%%%%%%%%%%%%%%%%
% disturbances 
%valve opening [%] 
cv101 = P_vector(1);
cv102 = P_vector(2);
cv103 = P_vector(3);
% if value is [A] from 0.004 to 0.020
% if you want to convert to 0 (fully closed) to 1 (fully open)
% vo_n = (vo - 0.004)./(0.02 - 0.004);

%pump rotation [%]
pRate = P_vector(4);
% if value is [A] from 0.004 to 0.020
% if you want to convert to (min speed - max speed)
% goes from 12% of the max speed to 92% of the max speed
% pRate = 12 + (92 - 12)*(P_vector(4) - 0.004)./(0.02 - 0.004);

% always maintain the inputs greater than 0.5
% inputs computed in the previous MPC iteration
% Note that the inputs are the setpoints to the gas flowrate PID's
fic104sp = P_vector(5);
fic105sp = P_vector(6); 
fic106sp = P_vector(7);
%current inputs of the plant
u0old=[P_vector(5);P_vector(6);P_vector(7)];


% cropping the data vector
nd = size(I_vector,2);
dataCrop = (nd - BufferLength + 1):nd;

% liquid flowrates [L/min] %INCLUDE NOISE???
fi101 = I_vector(1,dataCrop);
fi102 = I_vector(2,dataCrop);
fi103 = I_vector(3,dataCrop);

% actual gas flowrates [sL/min] %INCLUDE NOISE???
fic104 = I_vector(4,dataCrop);
fic105 = I_vector(5,dataCrop);
fic106 = I_vector(6,dataCrop);

% pressure @ injection point [mbar g]
pi105 = I_vector(7,dataCrop);
pi106 = I_vector(8,dataCrop);
pi107 = I_vector(9,dataCrop);

% reservoir outlet temperature [oC]
ti101 = I_vector(10,dataCrop);
ti102 = I_vector(11,dataCrop);
ti103 = I_vector(12,dataCrop);

% DP @ erosion boxes [mbar]
dp101 = I_vector(13,dataCrop);
dp102 = I_vector(14,dataCrop);
dp103 = I_vector(15,dataCrop);

% top pressure [mbar g]
% for conversion [bar a]-->[mbar g]
% ptop_n = ptop*10^-3 + 1.01325;
pi101 = I_vector(16,dataCrop);
pi102 = I_vector(17,dataCrop);
pi103 = I_vector(18,dataCrop);

% reservoir pressure [bar g]
% for conversion [bar g]-->[bar a]
% ptop_n = ptop + 1.01325;
pi104 = I_vector(19,dataCrop);

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CODE GOES HERE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% number of measurements in the data window
dss = size(pi104,2);

% we do no consider SS detection here
flagSS = 1;

% check for first iteration
if ~exist('dxHatkk','var') 
    %initial condition
    [dx0,z0,u0,theta0] = InitialConditionGasLift(par);

    dxHatkk = dx0;
    zHatkk = z0;
    thetakk = theta0;

    load('EKFconf');   
    qThres = ekf.qThresk;
    Pkk = ekf.Pkk;
%     ekf.R = noise.output; %added (different from Rig Implementation)
    
    lambda = 9.5;
    lambdaOrg = 9.5;
    lambda_hat = 9.5;
    Total_uOpt = 7.5;
    udirtot = 7.5;
    gainSwitch = 0;
    uOpt = u0old;
    uDir = u0old;
    
    err0 = zeros(3,1);
end

if flagSS == 1
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Estimating SS model parameters (dynamic) %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    yPlant = [fi101; 
              fi102;
              fi103;
              1.01325 + 10^-3*pi101; %[mbarg]-->[bar a];
              1.01325 + 10^-3*pi102;
              1.01325 + 10^-3*pi103];
          
    uPlant = [fic104; %conversion [L/min] --> [kg/s]
              fic105;
              fic106;
              cv101*ones(1,dss); %workaround - i just have the last measurement here. Since it is the disturbance, it doesn't really matter;
              cv102*ones(1,dss);
              cv103*ones(1,dss);
              pi104 + 1.01325];            
    
          try
              % running casadi - getting the last measurement and last
              % input that generated that measuremnt
              [dxHatkk,zHatkk,thetakk,Pkk,qThres] = RExtendedKalmanFilter(yPlant(:,end),uPlant(:,end - 10),dxHatkk,zHatkk,thetakk,Pkk,F_model,S_xx,S_zz,S_xz,S_xp,S_zp,ekf,par);
              
              % everything normal
              flagEst = 1;
          catch
              warning('Dynamic Estimation Problem!');
              beep
              % estimation problem
              flagEst = 0;
              
              % Note that in this case, we dont update
              % dxHatkk,zHatkk,thetakk,Pkk,qThres
          end
    
    
    if flagEst == 1
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Optimizing Systems %
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % MV Selector Matrix
        Hc = [0; 0; 1]';
        
        %--------------------------------%
        % Central Constraint Controller  %
        %--------------------------------%
        
        % Update Lambda
        Total_Qgl = fic104(end) + fic105(end) + fic106(end);
        K_AW = 1/1;
        %K_ccc = 1.621;
%         K_ccc = 0.86967;
        K_ccc = 0.2171164;

%         tau_ccc = 50;
        tau1_ccc = 48;
        tau_ccc = 125;
        theta_ccc = 0;
 
        KP_ccc = tau1_ccc/(K_ccc*(tau_ccc+theta_ccc));
        KI_ccc = 1/(K_ccc*(tau_ccc+theta_ccc));
        % lambda_next =  max(0,lambda + 0.75 * (Total_Qgl   - 7.5));
        
        tStart = cputime;
        
        lambda_next =  lambda + KI_ccc*(Hc*uOpt - Hc*uDir) + gainSwitch*KP_ccc*(Hc*uOpt - Hc*uDir);% ;% + K_AW*(lambda - lambda_hat); 
%         lambda_next = lambda;
        lambda_hat = lambda_next;

        lambda_next = max(0,lambda_hat);
        
        tEnd_lambda = cputime - tStart;
        
        lambda = lambda_next;
        lambdaOrg =  max(0,lambda + KI_ccc*(Total_Qgl   - 7.5)); 

%          if kk>500
%             lambda = 10.5;
%          else
%              lambda = 9.5;
%          end
        
        
        % Gradient Estimation
        %conversion
        CR = 60*10^3; % [L/min] -> [m3/s] 
        
        %%% NEED TO CHECK
        F_zp = full(S_zp(dxHatkk,zHatkk,[uPlant(:,end);thetakk;1e4]));
        
        tStart = cputime;
        %F_zp = full(S_zp(dxHatkk,zHatkk,[uSPPlantArray(1:3,end);uPlant(4:end,end);thetakk;1e4]));
        tEnd_lambda = cputime - tStart + tEnd_lambda;
        %%% NEED TO CHECK

        J_qgl1 = - 20*((1*1e-2)*CR/par.rho_o(1))*F_zp(22,1);
        J_qgl2 = - 25*((1*1e-2)*CR/par.rho_o(2))*F_zp(23,2);
        J_qgl3 = - 30*((1*1e-2)*CR/par.rho_o(3))*F_zp(24,3);     

        g_qgl1 = 1;
        g_qgl2 = 1;
        g_qgl3 = 1;
        
        L_qgl1 = J_qgl1 + lambda.*g_qgl1;
        L_qgl2 = J_qgl2 + lambda.*g_qgl2;
        L_qgl3 = J_qgl3 + lambda.*g_qgl3;
        
        %Lu(:,kk) = [L_qgl1;L_qgl2;L_qgl3];
        
        tauC = 10;
        tauI = min(4*tauC,800);
        
       
        % --------------------------------- %
        % ----- Direct Const. Control ----- %
        % --------------------------------- %
        taui_dc = 2.5;
        K_dir = 1;
        tauc_dc = 10;
        KI_dir = 1/(K_dir*tauc_dc);
        KP_dir = taui_dc/((K_dir*tauc_dc));
        K_aw = -KI_dir/KP_dir;
%        K_dir = 1;

        udir1 = max(1,min(5.5, uDir(1) + Hc(1)*KI_dir*(7.5 - Total_Qgl)+ K_aw*(uDir(1)-u0old(1))));
        udir2 = max(1,min(5.5, uDir(2) + Hc(2)*KI_dir*(7.5 - Total_Qgl)+ K_aw*(uDir(2)-u0old(2))));
        udir3 = max(1,min(5.5, uDir(3) + Hc(3)*KI_dir*(7.5 - Total_Qgl)+ K_aw*(uDir(3)-u0old(3))));%
        %udir3 = max(1,min(5.5, u0old(3) + Hc(3)*KI_dir*(7.5 - Total_Qgl)+ K_aw*(uDir(3)-u0old(3))));%

            
        udirtot = udir1 + udir2 + udir3;    
        uDir = [udir1; udir2; udir3];
        udir(:,i) = [udir1; udir2; udir3; udirtot];

        % --------------------------------- %
        % ----- Gradient Controllers ------ %
        % --------------------------------- %
        
        % ------------ Well 1 ------------- %

        K_1 = 2.53;

        tauC_1 = 25;
        theta_1 = 0;
        Ki1 = 1/(K_1*tauC_1 + theta_1);

        err_1= -L_qgl1;
        
        tStart = cputime;
                uOpt(1,1) = max(1,min(5.5, u0old(1) + (Ki1*err_1)));
        tEnd_gr1 = cputime - tStart;        
        
        err0(1) = err_1;
        
        % ------------ Well 2 ------------- %

        K_2 = 2.93;

        tauC_2 = 25;
        theta_2 = 0;
        Ki2 = 1/(K_2*tauC_2 + theta_2);
        err_2 = -L_qgl2;
        
            tStart = cputime;
                uOpt(2,1) = max(1,min(5.5, u0old(2) + (Ki2*err_2)));
            tEnd_gr2 = cputime - tStart; 

        err0(2) = err_2;
        
        % ------------ Well 3 ------------- %

        K_3 = 3.70;

        tauC_3 = 25;%14/2;
        theta_3 = 0;
        Ki3 = 1/(K_3*tauC_3 + theta_3);
        %Ki1 = Kp1/tauI;
        err_3= -L_qgl3;

            tStart = cputime;
                 uOpt(3,1) = max(1,min(5.5, u0old(3) + (Ki3*err_3)));
            tEnd_gr3 = cputime - tStart;

        err0(3) = err_3;        
        
%         uOpt=[u0old(1) u0old(2) u0old(3)]';
%         if kk > 500
%             uDir(3) = 2.0;
%         else
%             uDir(3) = u0old(3);
%         end
        flagOpt = 1;
        
        tEnd = max(max(tEnd_gr1,tEnd_gr2),tEnd_gr3)+ tEnd_lambda;
        % getting mean value of the inputs as initial guess for optimization
        %uGuess = mean(uPlant,2);
        %uGuess(1:3) = [2.5; 2.5; 2.5];
        Total_uOpt = uOpt(1) + uOpt(2) + uOpt(3);
        uind(:,i) = [uOpt(1);uOpt(2);uOpt(3);Total_uOpt];
        
        if uOpt(3)>uDir(3)
            indordir(:,i)=1;
            gainSwitch = 1;
        else
            indordir(:,i)=0;
            gainSwitch = 0;
        end
        
%         if Total_uOpt > 7.5
%             uOpt(3)=uDir(3);
%         end
        
        %[uOpt,xOptHat,zOptHat,phi,flagOpt] = ErosionRigSSOptimization(dxHatkk,zHatkk,thetakk,uGuess,par);

        if flagOpt == 1
            %%%%%%%%%%%%%%%%%%%%%%%%%%%
            %Send Variable
            %%%%%%%%%%%%%%%%%%%%%%%%%%%
            % compute new values for the gas flow rate setpoints
            %Filter for optimum
            fic104sp = uOpt(1);%min(uOpt(1),uDir(1));
            fic105sp = uOpt(2);%min(uOpt(2),uDir(2));
            fic106sp = min(uOpt(3),uDir(3));%uOpt(3);

            O_vector = vertcat(fic104sp,fic105sp,fic106sp)';
            SS = 1;
            Estimation = 1;
            Optimization = 1;
            Result = lambda; 
            Parameter_Estimation = thetakk';
            State_Variables_Estimation = (par.H*zHatkk)';
            State_Variables_Optimization = ([J_qgl1;J_qgl2;J_qgl3;L_qgl1;L_qgl2;L_qgl3;])';
            Optimized_Air_Injection = uOpt';
        else
            %%%%%%%%%
            %(dummy)%
            %%%%%%%%%
            % compute new values for the gas flow rate setpoints
            O_vector = vertcat(fic104sp,fic105sp,fic106sp)';

            SS = 1;
            Estimation = 1;
            Optimization = 0;
            Result = 0;
            Parameter_Estimation = [0,0,0,0,0,0];
            State_Variables_Estimation = [0,0,0,0,0,0];
            State_Variables_Optimization = [0,0,0,0,0,0];
            Optimized_Air_Injection = [0,0,0];
        end
        
    else  
        %%%%%%%%%
        %(dummy)%
        %%%%%%%%%
        % compute new values for the gas flow rate setpoints
        O_vector = vertcat(fic104sp,fic105sp,fic106sp)';

        SS = 1;
        Estimation = 0;
        Optimization = 0;
        Result = 0;
        Parameter_Estimation = [0,0,0,0,0,0];
        State_Variables_Estimation = [0,0,0,0,0,0];
        State_Variables_Optimization = [0,0,0,0,0,0];
        Optimized_Air_Injection = [0,0,0];
    end
    
    
else
    %%%%%%%%%
    %(dummy)%
    %%%%%%%%%
    % compute new values for the gas flow rate setpoints
    O_vector = vertcat(fic104sp,fic105sp,fic106sp)';
    
    SS = 0;
    Estimation = 0;
    Optimization = 0;
    Result = 0;
    Parameter_Estimation = [0,0,0,0,0,0];
    State_Variables_Estimation = [0,0,0,0,0,0];
    State_Variables_Optimization = [0,0,0,0,0,0];
    Optimized_Air_Injection = [0,0,0];
    
end
